import torch
from gymnasium import spaces
from stable_baselines3 import PPO, SAC


class OnnxablePolicy(torch.nn.Module):
    def __init__(
        self,
        obs_keys=None,
        features_extractor=None,
        mlp_extractor=None,
        action_net=None,
        value_net=None,
        use_obs_array=None,
        actor=None,
    ):
        super().__init__()
        self.obs_keys = obs_keys
        self.features_extractor = features_extractor
        self.mlp_extractor = mlp_extractor
        self.action_net = action_net
        self.value_net = value_net
        self.use_obs_array = use_obs_array
        self.actor = actor

    def forward_sac(self, observation: torch.Tensor, state_ins):
        return self.actor(observation, deterministic=True), state_ins

    def forward_ppo(self, obs, state_ins):
        features = None

        if self.use_obs_array:
            features = self.features_extractor(obs)
        else:
            obs_dict = {k: v for k, v in zip(self.obs_keys, obs)}
            features = self.features_extractor(obs_dict)

        action_hidden, value_hidden = self.mlp_extractor(features)
        return self.action_net(action_hidden), state_ins

    def forward(self, obs, state_ins):
        if self.actor:
            return self.forward_sac(obs, state_ins)
        else:
            return self.forward_ppo(obs, state_ins)


def export_model_as_onnx(model, onnx_model_path: str, use_obs_array: bool = False, obs_keys=None):
    if obs_keys is None:
        obs_keys = ["obs"]
    policy = model.policy.to("cpu")
    dummy_input = None
    onnxable_model = None

    if isinstance(model, SAC):
        assert use_obs_array, "SAC ONNX export works with use_obs_array=True, MLPPolicy and SBGSingleObsEnv only."

    if isinstance(model, PPO):
        onnxable_model = OnnxablePolicy(
            obs_keys,
            policy.features_extractor,
            policy.mlp_extractor,
            policy.action_net,
            policy.value_net,
            use_obs_array,
        )
        if use_obs_array:
            dummy_input = torch.unsqueeze(torch.tensor(model.observation_space.sample()), 0)
        else:
            dummy_input = dict(model.observation_space.sample())
            for k, v in dummy_input.items():
                dummy_input[k] = torch.from_numpy(v).unsqueeze(0)
            dummy_input = [v for v in dummy_input.values()]

    elif isinstance(model, SAC):
        onnxable_model = OnnxablePolicy(actor=model.policy.actor)
        dummy_input = torch.randn(1, *model.observation_space.shape)

    torch.onnx.export(
        onnxable_model,
        args=(dummy_input, torch.zeros(1).float()),
        f=onnx_model_path,
        opset_version=17,
        input_names=["obs", "state_ins"],
        output_names=["output", "state_outs"],
        dynamic_axes={
            "obs": {0: "batch_size"},
            "state_ins": {0: "batch_size"},  # variable length axes
            "output": {0: "batch_size"},
            "state_outs": {0: "batch_size"},
        },
    )

    # We only verify with PPO currently due to different output shape with SAC
    # (this can be updated in the future)
    if isinstance(model, PPO):
        # If the space is MultiDiscrete, we skip verifying as action output will have an expected mismatch
        # (the output from onnx will be the action logits for each discrete action,
        # while the output from sb3 will be a single int)
        if not isinstance(model.action_space, spaces.MultiDiscrete):
            verify_onnx_export(model, onnx_model_path, use_obs_array=use_obs_array)


def verify_onnx_export(ppo: PPO, onnx_model_path: str, num_tests=10, use_obs_array: bool = False):
    import numpy as np
    import onnx
    import onnxruntime as ort

    onnx_model = onnx.load(onnx_model_path)
    onnx.checker.check_model(onnx_model)

    sb3_model = ppo.policy.to("cpu")
    ort_sess = ort.InferenceSession(onnx_model_path, providers=["CPUExecutionProvider"])

    for i in range(num_tests):
        obs = None
        obs2 = None

        if use_obs_array:
            obs = np.expand_dims(ppo.observation_space.sample(), axis=0)
            obs2 = torch.tensor(obs)
        else:
            obs = dict(ppo.observation_space.sample())
            obs2 = {}
            for k, v in obs.items():
                obs2[k] = torch.from_numpy(v).unsqueeze(0)
            obs = [v for v in obs.values()]

        with torch.no_grad():
            action_sb3, _, _ = sb3_model(obs2, deterministic=True)

        action_onnx, state_outs = ort_sess.run(None, {"obs": obs, "state_ins": np.array([0.0], dtype=np.float32)})
        assert np.allclose(action_sb3, action_onnx, atol=1e-5), "Mismatch in action output"
        assert np.allclose(state_outs, np.array([0.0]), atol=1e-5), "Mismatch in state_outs output"



"""
This is the main entrypoint to the Godot RL Agents interface

Example usage is best found in the documentation:
https://github.com/edbeeching/godot_rl_agents/blob/main/docs/EXAMPLE_ENVIRONMENTS.md

Hyperparameters and training algorithm can be defined in a .yaml file, see ppo_test.yaml as an example.


Interactive Training:
With the Godot editor open, type gdrl in the terminal to launch training and
then press PLAY in the Godot editor. Training can be stopped with CTRL+C or
by pressing STOP in the editor.


Training with an exported executable:

gdrl --env_path path/to/exported/executable ---config_path path/to/yaml/file


"""

import argparse
import warnings

try:
    from godot_rl.wrappers.ray_wrapper import rllib_training
except ImportError as e:
    error_message = str(e)

    def rllib_training(args, extras):
        print("Import error importing rllib. If you have not installed the package, try: pip install godot-rl[rllib]")
        print("Otherwise try fixing the error.", error_message)


try:
    from godot_rl.wrappers.stable_baselines_wrapper import stable_baselines_training
except ImportError as e:
    error_message = str(e)

    def stable_baselines_training(args, extras):
        print("Import error importing sb3. If you have not installed the package, try: pip install godot-rl[sb3]")
        print("Otherwise try fixing the error.", error_message)


try:
    from godot_rl.wrappers.sample_factory_wrapper import sample_factory_enjoy, sample_factory_training
except ImportError as e:
    error_message = str(e)

    def sample_factory_training(args, extras):
        print(
            "Import error importing sample-factory If you have not installed the package, try: pip install godot-rl[sf]"
        )
        print("Otherwise try fixing the error.", error_message)


def get_args():
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument(
        "--trainer", default="sb3", choices=["sb3", "sf", "rllib"], type=str, help="framework to use (rllib, sf, sb3)"
    )
    parser.add_argument("--env_path", default=None, type=str, help="Godot binary to use")
    parser.add_argument(
        "--config_file", default="ppo_test.yaml", type=str, help="The yaml config file [only for rllib]"
    )
    parser.add_argument("--restore", default=None, type=str, help="the location of a checkpoint to restore from")
    parser.add_argument("--eval", default=False, action="store_true", help="whether to eval the model")
    parser.add_argument("--speedup", default=1, type=int, help="whether to speed up the physics in the env")
    parser.add_argument("--export", default=False, action="store_true", help="wheter to export the model")
    parser.add_argument("--num_gpus", default=None, type=int, help="Number of GPUs to use [only for rllib]")
    parser.add_argument(
        "--experiment_dir",
        default=None,
        type=str,
        help="The name of the the experiment directory, in which the tensorboard logs are getting stored",
    )
    parser.add_argument(
        "--experiment_name",
        default="experiment",
        type=str,
        help="The name of the the experiment, which will be displayed in tensborboard",
    )
    parser.add_argument("--viz", default=False, action="store_true", help="Whether to visualize one process")
    parser.add_argument("--seed", default=0, type=int, help="seed of the experiment")

    args, extras = parser.parse_known_args()
    if args.experiment_dir is None:
        args.experiment_dir = f"logs/{args.trainer}"

    if args.trainer == "sf" and args.env_path is None:
        print("WARNING: the sample-factory intergration is not designed to run in interactive mode, export you game")

    return args, extras


def main():
    warnings.warn(
        "This use of gdrl is deprecated and will be removed in version 1.0, please refer to the examples in the github repo",
        DeprecationWarning,
    )

    args, extras = get_args()
    if args.trainer == "rllib":
        training_function = rllib_training
    elif args.trainer == "sb3":
        training_function = stable_baselines_training
    elif args.trainer == "sf":
        if args.eval:
            training_function = sample_factory_enjoy
        else:
            training_function = sample_factory_training
    else:
        raise NotImplementedError

    training_function(args, extras)


if __name__ == "__main__":
    main()



# we download examples from github and we save them in the examples folder

import os
import shutil
from zipfile import ZipFile

import wget

BRANCHES = {"4": "main", "3": "godot3.5"}

BASE_URL = "https://github.com/edbeeching/godot_rl_agents_examples"


def download_examples():
    # select branch
    print("Select Godot version:")
    for key in BRANCHES.keys():
        print(f"{key} : {BRANCHES[key]}")

    branch = input("Enter your choice: ")
    BRANCH = BRANCHES[branch]
    os.makedirs("examples", exist_ok=True)
    URL = f"{BASE_URL}/archive/refs/heads/{BRANCH}.zip"
    print(f"downloading examples from {URL}")
    wget.download(URL, out="")
    print()
    print("unzipping")
    with ZipFile(f"{BRANCH}.zip", "r") as zipObj:
        # Extract all the contents of zip file in different directory
        zipObj.extractall("examples/")
    print("cleaning up")
    os.remove(f"{BRANCH}.zip")
    print("moving files")
    for file in os.listdir(f"examples/godot_rl_agents_examples-{BRANCH}"):
        shutil.move(f"examples/godot_rl_agents_examples-{BRANCH}/{file}", "examples")
    os.rmdir(f"examples/godot_rl_agents_examples-{BRANCH}")



import os
import pathlib
from typing import List, Optional, Tuple

import numpy as np
import ray
import yaml
from ray import tune
from ray.rllib.env.vector_env import VectorEnv
from ray.rllib.utils.typing import EnvActionType, EnvInfoDict, EnvObsType

from godot_rl.core.godot_env import GodotEnv


class RayVectorGodotEnv(VectorEnv):
    def __init__(self, port=10008, seed=0, config=None) -> None:
        config = config or {}  # initialize config as empty dict if None
        extra_arguments = {
            key: value
            for key, value in config.items()
            if key not in ["env_path", "show_window", "action_repeat", "speedup", "seed", "port"]
        }

        self._env = GodotEnv(
            env_path=config["env_path"],
            port=port,
            seed=seed,
            show_window=config["show_window"],
            action_repeat=config["action_repeat"],
            speedup=config["speedup"],
            **extra_arguments,
        )
        super().__init__(
            observation_space=self._env.observation_space,
            action_space=self._env.action_space,
            num_envs=self._env.num_envs,
        )

    def vector_reset(
        self, *, seeds: Optional[List[int]] = None, options: Optional[List[dict]] = None
    ) -> List[EnvObsType]:
        self.obs, info = self._env.reset()
        return self.obs, info

    def vector_step(
        self, actions: List[EnvActionType]
    ) -> Tuple[List[EnvObsType], List[float], List[bool], List[EnvInfoDict]]:
        actions = np.array(actions, dtype=np.dtype(object))
        self.obs, reward, term, trunc, info = self._env.step(actions, order_ij=True)
        return self.obs, reward, term, trunc, info

    def get_unwrapped(self):
        return [self._env]

    def reset_at(
        self,
        index: Optional[int] = None,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> EnvObsType:
        # the env is reset automatically, no need to reset it
        return self.obs[index], {}


def register_env():
    tune.register_env(
        "godot",
        lambda c: RayVectorGodotEnv(
            config=c,
            port=c.worker_index + GodotEnv.DEFAULT_PORT + 10,
            seed=c.worker_index + c["seed"],
        ),
    )


# Refactored section: Commented onnx section was removed as it was re-implemented in rllib_example.py


def rllib_training(args, extras):
    with open(args.config_file) as f:
        exp = yaml.safe_load(f)
    register_env()

    exp["config"]["env_config"]["env_path"] = args.env_path
    exp["config"]["env_config"]["seed"] = args.seed

    if args.env_path is not None:
        run_name = exp["algorithm"] + "/" + pathlib.Path(args.env_path).stem
    else:
        run_name = exp["algorithm"] + "/editor"
    print("run_name", run_name)

    if args.num_gpus is not None:
        exp["config"]["num_gpus"] = args.num_gpus

    if args.env_path is None:
        print("SETTING WORKERS TO 1")
        exp["config"]["num_workers"] = 1

    checkpoint_freq = 10

    exp["config"]["env_config"]["show_window"] = args.viz
    exp["config"]["env_config"]["speedup"] = args.speedup

    if args.eval or args.export:
        checkpoint_freq = 0
        exp["config"]["env_config"]["show_window"] = True
        exp["config"]["env_config"]["framerate"] = None
        exp["config"]["lr"] = 0.0
        exp["config"]["num_sgd_iter"] = 1
        exp["config"]["num_workers"] = 1
        exp["config"]["train_batch_size"] = 8192
        exp["config"]["sgd_minibatch_size"] = 128

        exp["config"]["explore"] = False
        exp["stop"]["training_iteration"] = 999999

    print(exp)

    ray.init(num_gpus=exp["config"]["num_gpus"] or 1)

    if not args.export:
        tune.run(
            exp["algorithm"],
            name=run_name,
            config=exp["config"],
            stop=exp["stop"],
            verbose=3,
            checkpoint_freq=checkpoint_freq,
            checkpoint_at_end=not args.eval,
            restore=args.restore,
            storage_path=os.path.abspath(args.experiment_dir) or os.path.abspath("logs/rllib"),
            trial_name_creator=lambda trial: (
                f"{args.experiment_name}" if args.experiment_name else f"{trial.trainable_name}_{trial.trial_id}"
            ),
        )
    if args.export:
        raise NotImplementedError("Use examples/rllib_example.py to export to onnx.")
        # rllib_export(args.restore)

    ray.shutdown()



from typing import Any, Dict, List, Tuple

import gymnasium as gym
import numpy as np

from godot_rl.wrappers.stable_baselines_wrapper import StableBaselinesGodotEnv

# A variant of the Stable Baselines Godot Env that only supports a single obs space from the dictionary - obs["obs"] by default.
# This provides some basic support for using envs that have a single obs space with policies other than MultiInputPolicy.


class SBGSingleObsEnv(StableBaselinesGodotEnv):
    def __init__(self, obs_key="obs", *args, **kwargs) -> None:
        self.obs_key = obs_key
        super().__init__(*args, **kwargs)

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[Dict[str, Any]]]:
        obs, rewards, term, info = super().step(action)

        # Terminal obs info is needed for imitation learning
        for idx, done in enumerate(term):
            if done:
                info[idx]["terminal_observation"] = obs[self.obs_key][idx]

        return obs[self.obs_key], rewards, term, info

    def reset(self) -> np.ndarray:
        obs = super().reset()
        return obs[self.obs_key]

    @property
    def observation_space(self) -> gym.Space:
        return self.envs[0].observation_space[self.obs_key]



# PettingZoo wrapper for GDRL
# Multi-agent, where 1 agent corresponds to one AIController instance in Godot
# Based on https://pettingzoo.farama.org/content/environment_creation/#example-custom-parallel-environment
# https://github.com/Farama-Foundation/PettingZoo/?tab=License-1-ov-file#readme
# and adjusted to work with GodotRL and Rllib (made for and tested only with rllib for now)

import functools
from typing import Dict

import numpy as np
from pettingzoo import ParallelEnv

from godot_rl.core.godot_env import GodotEnv


def env(render_mode=None):
    """
    The env function often wraps the environment in wrappers by default.
    You can find full documentation for these methods
    elsewhere in the developer documentation.
    """
    # Not implemented
    return env


class GDRLPettingZooEnv(ParallelEnv):
    metadata = {"render_modes": ["human"], "name": "GDRLPettingZooEnv"}

    def __init__(self, port=GodotEnv.DEFAULT_PORT, show_window=True, seed=0, config: Dict = None):
        """
        The init method takes in environment arguments and should define the following attributes:
        - possible_agents
        - render_mode

        Note: as of v1.18.1, the action_spaces and observation_spaces attributes are deprecated.
        Spaces should be defined in the action_space() and observation_space() methods.
        If these methods are not overridden, spaces will be inferred from self.observation_spaces/action_spaces, raising a warning.

        These attributes should not be changed after initialization.
        """
        config = config or {}  # initialize config as empty dict if None
        extra_arguments = {
            key: value
            for key, value in config.items()
            if key not in ["env_path", "show_window", "action_repeat", "speedup", "seed", "port"]
        }

        # Initialize the Godot Env which we will wrap
        self.godot_env = GodotEnv(
            env_path=config["env_path"],
            show_window=show_window,
            action_repeat=config["action_repeat"],
            speedup=config["speedup"],
            convert_action_space=False,
            seed=seed,
            port=port,
            **extra_arguments,
        )

        self.render_mode = None  # Controlled by the env

        self.possible_agents = [agent_idx for agent_idx in range(self.godot_env.num_envs)]
        self.agents = self.possible_agents[:]

        # The policy names here are set on each AIController in Godot editor,
        # used to map agents to policies for multi-policy training.
        self.agent_policy_names = self.godot_env.agent_policy_names

        # optional: a mapping between agent name and ID
        self.agent_name_mapping = dict(zip(self.possible_agents, list(range(len(self.possible_agents)))))

        self.observation_spaces = {
            agent: self.godot_env.observation_spaces[agent_idx] for agent_idx, agent in enumerate(self.agents)
        }

        self.action_spaces = {
            agent: self.godot_env.tuple_action_spaces[agent_idx] for agent_idx, agent in enumerate(self.agents)
        }

    # Observation space should be defined here.
    # lru_cache allows observation and action spaces to be memoized, reducing clock cycles required to get each agent's space.
    # If your spaces change over time, remove this line (disable caching).
    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        return self.observation_spaces[agent]

    # Action space should be defined here.
    # If your spaces change over time, remove this line (disable caching).
    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        return self.action_spaces[agent]

    def render(self):
        """
        Renders the environment. In human mode, it can print to terminal, open
        up a graphical window, or open up some other display that a human can see and understand.
        """
        # Not implemented

    def close(self):
        """
        Close should release any graphical displays, subprocesses, network connections
        or any other environment data which should not be kept around after the
        user is no longer using the environment.
        """
        self.godot_env.close()

    def reset(self, seed=None, options=None):
        """
        Reset needs to initialize the `agents` attribute and must set up the
        environment so that render(), and step() can be called without issues.
        Returns the observations for each agent
        """
        godot_obs, godot_infos = self.godot_env.reset()

        observations = {agent: godot_obs[agent_idx] for agent_idx, agent in enumerate(self.agents)}
        infos = {agent: godot_infos[agent_idx] for agent_idx, agent in enumerate(self.agents)}

        return observations, infos

    def step(self, actions):
        """
        step(action) takes in an action for each agent and should return the
        - observations
        - rewards
        - terminations
        - truncations
        - infos
        dicts where each dict looks like {agent_1: item_1, agent_2: item_2}
        """

        # Once an agent (AIController) has done = true, it will not receive any more actions until all agents in the
        # Godot env have done = true. For agents that received no actions, we will set zeros instead for
        # compatibility.
        godot_actions = [
            (actions[agent] if agent in actions else np.zeros_like(self.action_spaces[agent_idx].sample()))
            for agent_idx, agent in enumerate(self.agents)
        ]

        godot_obs, godot_rewards, godot_dones, godot_truncations, godot_infos = self.godot_env.step(
            godot_actions, order_ij=True
        )
        observations = {agent: godot_obs[agent] for agent in actions}
        rewards = {agent: godot_rewards[agent] for agent in actions}

        terminations = {agent: godot_dones[agent] for agent in actions}

        # Truncations are not yet implemented in GDRL API
        truncations = {agent: False for agent in actions}

        # typically there won't be any information in the infos, but there must
        # still be an entry for each agent
        infos = {agent: godot_infos[agent] for agent in actions}

        return observations, rewards, terminations, truncations, infos
